/*
 * deviced
 *
 * Copyright (c) 2016 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


#include <stdio.h>
#include <stdbool.h>
#include <Ecore.h>
#include <hw/thermal.h>

#include "apps/apps.h"
#include "core/devices.h"
#include "core/log.h"
#include "core/device-notifier.h"

static struct thermal_device *thermal_dev;

static int noti;

#define ARRAY_SIZE(name) (sizeof(name)/sizeof(name[0]))

enum thermal_action_type {
	NO_CHANGE,
	LOW_TO_HIGH,
	HIGH_TO_LOW,
	HIGH_TO_CRITICAL,
	NEED_TO_CHECK
};

enum thermal_level_update_type {
	UPDATE,
	KEEP
};

struct thermal_state {
	thermal_state_e old_state;
	thermal_state_e new_state;
	enum thermal_action_type status;
};

struct thermal_level {
	thermal_level_e old_level;
	thermal_level_e new_level;
	enum thermal_action_type status;
	enum thermal_level_update_type level;
};

static struct thermal_state temp_state[] = {
	{ THERMAL_STATE_LOW,	THERMAL_STATE_LOW,	NO_CHANGE},
	{ THERMAL_STATE_LOW,	THERMAL_STATE_NORMAL,	NO_CHANGE},
	{ THERMAL_STATE_LOW,	THERMAL_STATE_HIGH,	LOW_TO_HIGH},
	{ THERMAL_STATE_NORMAL, THERMAL_STATE_LOW,	NO_CHANGE},
	{ THERMAL_STATE_NORMAL, THERMAL_STATE_NORMAL,	NO_CHANGE},
	{ THERMAL_STATE_NORMAL, THERMAL_STATE_HIGH,	LOW_TO_HIGH},
	{ THERMAL_STATE_HIGH,	THERMAL_STATE_LOW,	NEED_TO_CHECK},
	{ THERMAL_STATE_HIGH,	THERMAL_STATE_NORMAL,	NEED_TO_CHECK},
	{ THERMAL_STATE_HIGH,	THERMAL_STATE_HIGH,	NEED_TO_CHECK}
};

static struct thermal_level temp_level[] = {
	{ THERMAL_LEVEL_WARNING,	THERMAL_LEVEL_NORMAL,	HIGH_TO_LOW,		UPDATE},
	{ THERMAL_LEVEL_WARNING,	THERMAL_LEVEL_WARNING,	NO_CHANGE,		UPDATE},
	{ THERMAL_LEVEL_WARNING,	THERMAL_LEVEL_CRITICAL,	HIGH_TO_CRITICAL,	UPDATE},
	{ THERMAL_LEVEL_WARNING,	THERMAL_LEVEL_POWEROFF,	HIGH_TO_CRITICAL,	UPDATE},
	{ THERMAL_LEVEL_CRITICAL,	THERMAL_LEVEL_NORMAL,	NO_CHANGE,		KEEP},
	{ THERMAL_LEVEL_CRITICAL,	THERMAL_LEVEL_WARNING,	NO_CHANGE,		KEEP},
	{ THERMAL_LEVEL_CRITICAL,	THERMAL_LEVEL_CRITICAL,	NO_CHANGE,		UPDATE},
	{ THERMAL_LEVEL_CRITICAL,	THERMAL_LEVEL_POWEROFF,	NO_CHANGE,		UPDATE},
	{ THERMAL_LEVEL_POWEROFF,	THERMAL_LEVEL_NORMAL,	NO_CHANGE,		KEEP},
	{ THERMAL_LEVEL_POWEROFF,	THERMAL_LEVEL_WARNING,	NO_CHANGE,		KEEP},
	{ THERMAL_LEVEL_POWEROFF,	THERMAL_LEVEL_CRITICAL,	NO_CHANGE,		KEEP},
	{ THERMAL_LEVEL_POWEROFF,	THERMAL_LEVEL_POWEROFF,	NO_CHANGE,		UPDATE}
};

static int thermal_state_check(int old_state, int new_state)
{
	int i, ret = -1;

	for (i = 0; i < ARRAY_SIZE(temp_state); i++) {
		if (old_state == temp_state[i].old_state &&
			new_state == temp_state[i].new_state) {
			ret = temp_state[i].status;
			break;
		}
	}

	return ret;
}

static void thermal_level_check(int old_level, int new_level, int *out_status, int *out_level)
{
	int i;

	for (i = 0; i < ARRAY_SIZE(temp_level); i++) {
		if (old_level == temp_level[i].old_level &&
			new_level == temp_level[i].new_level) {
			*out_status = temp_level[i].status;
			*out_level = temp_level[i].level;
			break;
		}
	}
}

static void thermal_remove_noti(void)
{
	if (noti) {
		remove_notification("TempCooldownNotiOff", noti);
		noti = 0;
	}
}

static void thermal_remove_popup(void)
{
	launch_system_app(APP_DEFAULT, 2,
		APP_KEY_TYPE, "remove_cooldown_popups");
}

static void thermal_add_noti(void)
{
	if (!noti)
		noti = add_notification("TempCooldownNotiOn");
}

static void thermal_add_popup(void)
{
	launch_system_app(APP_OVERHEAT, 2,
			APP_KEY_TYPE, "overheat");
}

static void thermal_add_recovery_popup(void)
{
	launch_system_app(APP_DEFAULT, 2,
		APP_KEY_TYPE, "cooled_down");
}

static void thermal_action(struct thermal_info *info, void *data)
{
	static int old_state = -1, old_level = -1;
	static int status, high_status, update;

	if (!info)
		return;

	if (old_state < 0) {
		old_state = info->state;
		old_level = info->level;
	}

	status = thermal_state_check(old_state, info->state);

	if (status == NO_CHANGE)
		update = UPDATE;
	else if (status == LOW_TO_HIGH) {
		thermal_remove_noti();
		thermal_remove_popup();
		thermal_add_noti();
		update = UPDATE;
	} else if (status == NEED_TO_CHECK) {
		thermal_level_check(old_level, info->level, &high_status, &update);
		_I("high_status = %d update = %d", high_status, update);
		if (high_status == HIGH_TO_LOW) {
			thermal_remove_popup();
			thermal_remove_noti();
			thermal_add_recovery_popup();
		} else if (high_status == HIGH_TO_CRITICAL) {
			thermal_remove_noti();
			thermal_remove_popup();
			thermal_add_popup();
		}
	}

	if (update == UPDATE) {
		old_state = info->state;
		old_level = info->level;
	}

	_I("Last old_state = %d old_level=%d", old_state, old_level);
}

static int thermal_service_start(void)
{
	int ret;

	if (!thermal_dev) {
		_E("Thermal service is not supported");
		return -ENOTSUP;
	}

	ret = thermal_dev->register_changed_event(thermal_action, NULL);
	if (ret < 0)
		_E("Failed to register thermal event (%d)", ret);

	return ret;
}

static int thermal_service_stop(void)
{
	int ret;

	if (!thermal_dev) {
		_E("Thermal service is not supported");
		return -ENOTSUP;
	}

	ret = thermal_dev->unregister_changed_event(thermal_action);
	if (ret < 0)
		_E("Failed to unregister thermal event (%d)", ret);

	return ret;
}

static int booting_done(void *data)
{
	static int done;

	if (data == NULL)
		return done;
	done = *(int *)data;
	if (done == 0)
		return done;

	_I("booting done");
	thermal_service_start();

	return done;
}

static void thermal_init(void *data)
{
	register_notifier(DEVICE_NOTIFIER_BOOTING_DONE, booting_done);
}

static void thermal_exit(void *data)
{
	unregister_notifier(DEVICE_NOTIFIER_BOOTING_DONE, booting_done);
	thermal_service_stop();
}

static int thermal_probe(void *data)
{
	struct hw_info *info;
	int ret;

	if (thermal_dev)
		return 0;

	ret = hw_get_info(THERMAL_HARDWARE_DEVICE_ID,
			(const struct hw_info **)&info);
	if (ret < 0) {
		_E("There is no HAL for Thermal");
		return ret;
	}

	if (!info->open) {
		_E("Failed to open thermal device; open(NULL)");
		return -ENODEV;
	}

	ret = info->open(info, NULL, (struct hw_common**)&thermal_dev);
	if (ret < 0) {
		_E("Failed to get thermal device structure (%d)", ret);
		return ret;
	}

	if (!thermal_dev ||
		!thermal_dev->get_state ||
		!thermal_dev->register_changed_event ||
		!thermal_dev->unregister_changed_event) {
		_E("Thermal HAL does not support event handler");
		return -ENODEV;
	}

	_I("Thermal device structure load success");
	return 0;
}

static const struct device_ops thermal_ops = {
	.name     = "thermal",
	.probe    = thermal_probe,
	.init     = thermal_init,
	.exit     = thermal_exit,
};

DEVICE_OPS_REGISTER(&thermal_ops)
